using System;
using System.Data;
using System.Collections.Generic;
using System.Text.RegularExpressions;
using SubSonic.Utilities;

namespace SubSonic
{
    public class StoredProcedure
    {
        public class Parameter
        {
            private DbType dbType;

            public DbType DBType
            {
                get { return dbType; }
                set { dbType = value; }
            }

	
            private string name;

            public string Name
            {
                get { return name; }
                set { name = value; }
            }

            private string displayName;

            public string DisplayName
            {
                get { return displayName; }
                set { displayName = value; }
            }

            private string queryParameter;

            public string QueryParameter
            {
                get { return queryParameter; }
                set { queryParameter = value; }
            }

            private ParameterDirection mode = ParameterDirection.Input;

            public ParameterDirection Mode
            {
                get { return mode; }
                set { mode = value; }
            }
        }

        public class ParameterCollection : List<Parameter>
        {
        }

        private ParameterCollection _parameters = new ParameterCollection();

        public ParameterCollection Parameters
        {
            get { return _parameters; }
            set { _parameters = value; }
        }

        private readonly DataProvider _provider;

        public DataProvider Provider
        {
            get { return _provider; }
        }

        public List<object> OutputValues;

        private QueryCommand _cmd;

        public QueryCommand Command
        {
            get { return _cmd; }
            set { _cmd = value; }
        }

        public IDataReader GetReader()
        {
            IDataReader result = DataService.GetReader(_cmd);
            OutputValues = _cmd.OutputValues;
            return result;
        }

        /// <summary>
        /// Executes an SP using the default provider and returns a reader
        /// </summary>
        /// <param name="spAndParams">The SP name and each argument required - these must be in order</param>
        /// <returns>System.Data.IDataReader</returns>
        public static IDataReader GetReader(string spAndParams)
        {
            return GetReader(spAndParams, "");
        }

        public static IDataReader GetReader(string spAndParams, string providerName) {
            QueryCommand cmd = new QueryCommand("EXEC " + spAndParams, providerName);
            cmd.CommandType = CommandType.Text;
            cmd.ProviderName = providerName;

            IDataReader result = DataService.GetReader(cmd);
            return result;
        }

        /// <summary>
        /// Executes an SP using the default provider
        /// </summary>
        /// <param name="spAndParams">The SP name and each argument required - these must be in order</param>
        /// <returns>System.Data.DataSet</returns>
        public static DataSet GetDataSet(string spAndParams)
        {
            return GetDataSet(spAndParams, String.Empty);
        }

        /// <summary>
        /// Executes an SP using the specified provider
        /// </summary>
        /// <param name="spAndParams">The SP name and each argument required - these must be in order</param>
		/// <param name="providerName">Specifies the specific provider to use.</param>
        /// <returns>System.Data.DataSet</returns>
        public static DataSet GetDataSet(string spAndParams, string providerName)
        {
            QueryCommand cmd = new QueryCommand("EXEC " + spAndParams, providerName);
            cmd.CommandType = CommandType.Text;
            cmd.ProviderName = providerName;
            return DataService.GetDataSet(cmd);
        }

        public DataSet GetDataSet()
        {
            return GetDataSet<DataSet>();
        }

        public T GetDataSet<T>() where T : DataSet, new()
        {
            T result = DataService.GetDataSet<T>(_cmd);
            OutputValues = _cmd.OutputValues;
            return result;
        }

		/// <summary>
		/// Executes the stored procedure and returns the number of rows affected.
		/// </summary>
		/// <returns></returns>
        public int Execute()
        {
            int rowsAffected = DataService.ExecuteQuery(_cmd);
            OutputValues = _cmd.OutputValues;
        	return rowsAffected;
        }

        public object ExecuteScalar()
        {
            object result = DataService.ExecuteScalar(_cmd);
            OutputValues = _cmd.OutputValues;
            return result;
        }

        private int commandTimeout = 60;

        /// <summary>
        /// Connection timeout in seconds
        /// </summary>
        public int CommandTimeout {
            get { return commandTimeout; }
            set { commandTimeout = value; }
        }

        private readonly string name;

        public string Name
        {
            get { return name; }
        }

        private readonly string displayName;

        public string DisplayName
        {
            get { return displayName; }
        }

        private readonly string tableName;

        public string TableName
        {
            get { return tableName; }
        }

        public StoredProcedure(string spName)
            : this(spName, DataService.Provider)
        {
        }

        public StoredProcedure(string spName, DataProvider provider)
        {
            name = spName;
            tableName = provider.SPClassName;
            _cmd = new QueryCommand(spName, provider.Name);
            _cmd.CommandType = CommandType.StoredProcedure;
            _provider = provider;
            _cmd.ProviderName = _provider.Name;
            _cmd.CommandTimeout = commandTimeout;
            string newName = spName;

            if(provider.ExtractClassNameFromSPName)
            {
                //Order of if/else matters here!
                Match underscoreMatch = Regex.Match(spName, RegexPattern.EMBEDDED_CLASS_NAME_UNDERSCORE_MATCH);
                if(underscoreMatch.Success)
                {
                    tableName = underscoreMatch.Value;
                    newName = Regex.Replace(spName, RegexPattern.EMBEDDED_CLASS_NAME_UNDERSCORE_REPLACE, String.Empty);
                }
                else
                {
                    Match match = Regex.Match(spName, RegexPattern.EMBEDDED_CLASS_NAME_MATCH);
                    if(match.Success)
                    {
                        tableName = match.Value;
                        newName = Regex.Replace(spName, RegexPattern.EMBEDDED_CLASS_NAME_REPLACE, String.Empty);
                    }
                }
            }

            displayName = TransformSPName(newName, provider);
            //init the list so the count comes back properly
            OutputValues = new List<object>();
        }

        private static string TransformSPName(string spName, DataProvider provider)
        {
            if(String.IsNullOrEmpty(spName))
            {
                return string.Empty;
            }

            string newName = spName;
            newName = Utility.RegexTransform(newName, provider);
            newName = Utility.StripText(newName, provider.StripSPText);
            newName = Utility.GetProperName(newName, provider);
            newName = Utility.IsStringNumeric(newName) ? "_" + newName : newName;
            newName = Utility.StripNonAlphaNumeric(newName);
            newName = newName.Trim();
            return Utility.KeyWordCheck(newName, String.Empty, provider);
        }
    }
}