using System;
using System.Collections.Generic;
using System.CodeDom;
using System.Text;
using SubSonic.CodeGenerator;
using System.Text.RegularExpressions;
using System.IO;
using SubSonic.Utilities;
using System.Collections.Specialized;

namespace SubSonic
{
    public static class CodeService
    {

        #region Helpers
        private static string templateDirectory = string.Empty;
        public static string TemplateDirectory
        {
            get { return templateDirectory; }
            set { templateDirectory = value; }
        }
        public enum TemplateType
        {
            Class,
            ODSController,
            ReadOnly,
            SP,
            Structs,
            DynamicScaffold,
            GeneratedScaffoldCodeBehind,
            GeneratedScaffoldMarkup
        }

        public enum ReplacementVariable
        {
            Table,
            Provider,
            View,
            StoredProcedure
        }
        public class Replacement
        {
            public Replacement(ReplacementVariable variable, string replace)
            {
                Variable = variable;
                ReplaceWith = replace;
            }

            private ReplacementVariable replaceVar;
            public ReplacementVariable Variable
            {
                get { return replaceVar; }
                set { replaceVar = value; }
            }


            private string replaceWith;
            public string ReplaceWith
            {
                get { return replaceWith; }
                set { replaceWith = value; }
            }

        }

        #endregion

        /// <summary>
        /// Not currently used, but will be the basis for user defined templates. Please don't remove!
        /// </summary>
        /// <param name="templateFile"></param>
        /// <param name="values"></param>
        /// <returns></returns>
        public static string RunTemplate(string templateFile, NameValueCollection values, ICodeLanguage language, DataProvider provider)
        {
            string templatePath = Path.Combine(TemplateDirectory, templateFile);
            string templateText = Sugar.Files.GetFileText(templatePath);

            for(int i = 0; i < values.Count; i++)
            {
                templateText = templateText.Replace(values.GetKey(i), values.Get(i));
            }

            TurboTemplate t = new TurboTemplate(templateText, language, provider);
            return t.Render();
        }

        public static TurboTemplate BuildTemplate(TemplateType templateType, NameValueCollection values, ICodeLanguage language, DataProvider provider)
        {
            string templateText = GetTemplateText(templateType, language);

            for (int i = 0; i < values.Count; i++)
            {
                templateText = templateText.Replace(values.GetKey(i), values.Get(i));
            }
            TurboTemplate t = new TurboTemplate(templateText, language, provider);
            return t;
        }

        public static string RunTemplate(TemplateType templateType, IEnumerable<Replacement> settings, ICodeLanguage language, DataProvider provider)
        {
            Utility.WriteTrace("Getting Template");
            string templateText = GetTemplateText(templateType, language);

            //set the provider and tablename
            Utility.WriteTrace("Replacing values in template");
            foreach(Replacement var in settings)
            {
                string replaceHolder = "#" + Enum.GetName(typeof(ReplacementVariable), var.Variable).ToUpper() + "#";
                templateText = Utility.FastReplace(templateText, replaceHolder, var.ReplaceWith, StringComparison.InvariantCultureIgnoreCase);
            }

            TurboTemplate t = new TurboTemplate(templateText, language, provider);
            Utility.WriteTrace("Rendering template");
            string output = t.Render();
            Utility.WriteTrace("Finished :)");
            return output;
        }

        public static bool ShouldGenerate(string objectName, string[] includeList, string[] excludeList, DataProvider provider)
        {
            bool result = true;
            bool generateAll = false;

            //first, check to see if the includeList says to include all tables
            //this is a default
            if (includeList.Length == 1)
            {
                if (includeList[0] == "*")
                {
                    generateAll = true;
                }
            }

            //if we need to generate all tables, then we need to check the excludeList
            if (generateAll)
            {
                foreach (string s in excludeList)
                {
                    if (Utility.IsRegexMatch(objectName, s.Trim()))
                    {
                        result = false;
                        break;
                    }
                }
            }
            else
            {
                //IncludeList TRUMPs excludeList in case of confusion
                //what this means is that if there is an includeList,
                //be definition there's an excludeList of all tables not included
                //yep, confusing.

                //this means that tables were specifically requested in the includeList
                //need to make them prove themselves
                result = false;

                foreach (string s in includeList)
                {
                    if (Utility.IsRegexMatch(objectName, s.Trim()))
                    {
                        result = true;
                        break;
                    }
                }
            }

            return result;
        }

        public static bool ShouldGenerate(TableSchema.Table tbl)
        {
            return ShouldGenerate(tbl.TableName, tbl.Provider.Name);
        }

        public static bool ShouldGenerate(string tableName, string providerName)
        {
            DataProvider provider = DataService.Providers[providerName];
            if (provider == null)
                throw new Exception("There is no provider with the name " + providerName);

            return ShouldGenerate(tableName, provider.IncludeTables, provider.ExcludeTables, provider);
        }



        public static TurboTemplate BuildODSTemplate(string tableName, ICodeLanguage language, DataProvider provider)
        {
            DataService.LoadProviders();

            if (ShouldGenerate(tableName, provider.Name) && provider.GenerateODSControllers)
            {
                List<Replacement> list = new List<Replacement>();
                list.Add(new Replacement(ReplacementVariable.Table, tableName));
                list.Add(new Replacement(ReplacementVariable.Provider, provider.Name));

                return PrepareTemplate("ODS Controller - " + provider.Name + ": " + tableName, TemplateType.ODSController, list, language, provider);
            }
            else
            {
                Utility.WriteTrace(tableName + " is Excluded from generation");
            }
            return null;
        }

        public static TurboTemplate BuildClassTemplate(string tableName, ICodeLanguage language, DataProvider provider)
        {
            DataService.LoadProviders();

            if (ShouldGenerate(tableName, provider.Name))
            {
                List<Replacement> list = new List<Replacement>();
                list.Add(new Replacement(ReplacementVariable.Table, tableName));
                list.Add(new Replacement(ReplacementVariable.Provider, provider.Name));

                return PrepareTemplate("Class - " + provider.Name + ": " + tableName, TemplateType.Class, list, language, provider);
            }
            else
            {
                Utility.WriteTrace("Class " + tableName + " is excluded from generation");
            }
            return null;
        }

        public static TurboTemplate BuildViewTemplate(string tableName, ICodeLanguage language, DataProvider provider)
        {
            DataService.LoadProviders();

            if (ShouldGenerate(tableName, provider.Name))
            {

                List<Replacement> list = new List<Replacement>();
                list.Add(new Replacement(ReplacementVariable.View, tableName));
                list.Add(new Replacement(ReplacementVariable.Provider, provider.Name));

                return PrepareTemplate("View - " + provider.Name + ": " + tableName, TemplateType.ReadOnly, list, language, provider);
            }
            else
            {
                Utility.WriteTrace("View " + tableName + " is excluded from generation");
            }
            return null;
        }

        public static TurboTemplate BuildSPTemplate(ICodeLanguage language, DataProvider provider)
        {
            DataService.LoadProviders();
            List<Replacement> list = new List<Replacement>();
            list.Add(new Replacement(ReplacementVariable.Provider, provider.Name));
            return PrepareTemplate("Stored Procedure Class - " + provider.Name, TemplateType.SP, list, language, provider);
        }

        public static TurboTemplate BuildStructsTemplate(ICodeLanguage language, DataProvider provider)
        {
            DataService.LoadProviders();
            List<Replacement> list = new List<Replacement>();
            return PrepareTemplate("Structs Class", TemplateType.Structs, list, language, provider);
        }


        static TurboTemplate PrepareTemplate(string templateName, TemplateType templateType, IEnumerable<Replacement> settings, ICodeLanguage language, DataProvider provider)
        {
            Utility.WriteTrace("Getting template for " + templateName);
            string templateText = GetTemplateText(templateType, language);
            //set the provider and tablename
            Utility.WriteTrace("Preparing template for " + templateName);
            foreach (Replacement var in settings)
            {
                string replaceHolder = "#" + Enum.GetName(typeof(ReplacementVariable), var.Variable).ToUpper() + "#";
                templateText = Utility.FastReplace(templateText, replaceHolder, var.ReplaceWith, StringComparison.InvariantCultureIgnoreCase);
            }
            TurboTemplate t = new TurboTemplate(templateText, language, provider);
            t.TemplateName = templateName;
            return t;
        }


        public static string RunODS(string tableName, ICodeLanguage language, DataProvider provider)
        {
            TurboTemplate tt = BuildODSTemplate(tableName, language, provider);
            return tt.Render();
        }

        public static string RunReadOnly(string viewName, ICodeLanguage language, DataProvider provider)
        {
            TurboTemplate tt = BuildViewTemplate(viewName, language, provider);
            return tt.Render();
        }

        public static string RunClass(string tableName, ICodeLanguage language, DataProvider provider)
        {
            TurboTemplate tt = BuildClassTemplate(tableName, language, provider);
            return tt.Render();
        }

        public static string RunSPs(ICodeLanguage language, DataProvider provider)
        {
            TurboTemplate tt = BuildSPTemplate(language, provider);
            return tt.Render();
        }

        public static string RunStructs(ICodeLanguage language, DataProvider provider)
        {
            TurboTemplate tt = BuildStructsTemplate(language, provider);
            return tt.Render();
        }

        public static CodeCompileUnit BuildCompilUnit(string code)
        {
            return new CodeSnippetCompileUnit(code);
        }

        static string GetTemplateText(TemplateType t, ICodeLanguage language)
        {
        	string template = language.TemplatePrefix;
            string templateText = null;

            switch (t)
            {
                case TemplateType.Class:
                    template += TemplateName.CLASS;
                    break;
                case TemplateType.ReadOnly:
                    template += TemplateName.VIEW;
                    break;
                case TemplateType.SP:
                    template += TemplateName.STORED_PROCEDURE;
                    break;
                case TemplateType.Structs:
                    template += TemplateName.STRUCTS;
                    break;
                case TemplateType.ODSController:
                    template += TemplateName.ODS_CONTROLLER;
                    break;
                case TemplateType.DynamicScaffold:
                    template += TemplateName.DYNAMIC_SCAFFOLD;
                    break;
                case TemplateType.GeneratedScaffoldCodeBehind:
                    template += TemplateName.GENERATED_SCAFFOLD_CODE_BEHIND;
                    break;
                case TemplateType.GeneratedScaffoldMarkup:
                    template += TemplateName.GENERATED_SCAFFOLD_MARKUP;
                    break;
                default:
                    template += TemplateName.CLASS;
                    break;
            }

        	template += FileExtension.DOT_ASPX;

			//decide where to pull the text from
            if (!String.IsNullOrEmpty(templateDirectory))
            {
                Utility.WriteTrace("Looking for template " + template + " in " + templateDirectory);

                //make sure the template exists
                string templatePath = Path.Combine(templateDirectory, template);

                if (File.Exists(templatePath))
                {
                    //pull the text from there
                    templateText = Sugar.Files.GetFileText(templatePath);
                }
                else
                {
                    Utility.WriteTrace("Template " + template + " NOT FOUND in directory " + templateDirectory + "; using embedded resource template instead...");
                }
            }

            if (String.IsNullOrEmpty(templateText))
            {
                Utility.WriteTrace("Loading template from resource: " + template);
                templateText = TurboTemplate.LoadTextFromManifest(template);
            }

            if (String.IsNullOrEmpty(templateText))
                throw new Exception("The template \"" + template + "\" is empty or cannot be found.");

            return templateText;
        }
    }
}
