using System;
using System.Text;
using System.Collections.Specialized;
using System.Reflection;
using System.CodeDom.Compiler;
using System.Text.RegularExpressions;
using SubSonic.Utilities;

namespace SubSonic.CodeGenerator
{
    public class TurboCompiler
    {
        private static readonly Regex regLineFix = new Regex(@"[\r\n]+", RegexOptions.Compiled);

        public void AddTemplate(TurboTemplate template)
        {
            if (templates.Count == 0)
            {
                References = template.References;
                Language = template.CompileLanguage;
            }
            template.EntryPoint = "Render";
            template.GeneratedRenderType = "Parser" + Templates.Count;
            template.TemplateText = Utility.FastReplace(template.TemplateText, "#TEMPLATENUMBER#", Templates.Count.ToString(), StringComparison.InvariantCultureIgnoreCase);
            templates.Add(template);
        }

        public void Reset()
        {
            templates.Clear();
        }

        private readonly TurboTemplateCollection templates = new TurboTemplateCollection();
        public TurboTemplateCollection Templates
        {
            get { return templates; }
        }

        private CodeDomProvider codeProvider = null;
        private CodeDomProvider CodeProvider
        {
            get
            {
                if (codeProvider == null)
                {
                    codeProvider = Language.CreateCodeProvider();
                }
                return codeProvider;
            }
        }

        private CompilerParameters codeCompilerParameters = null;
        private CompilerParameters CodeCompilerParameters
        {
            get
            {
                if (codeCompilerParameters == null)
                {
                    codeCompilerParameters = new CompilerParameters();
                    codeCompilerParameters.CompilerOptions = "/target:library"; // you can add /optimize
                    codeCompilerParameters.GenerateExecutable = false;
                    codeCompilerParameters.GenerateInMemory = true;
                    codeCompilerParameters.IncludeDebugInformation = false;
                    codeCompilerParameters.ReferencedAssemblies.Add("mscorlib.dll");
                    foreach (string s in References)
                    {
                        codeCompilerParameters.ReferencedAssemblies.Add(s);
                    }
                }
                return codeCompilerParameters;
            }
        }

        public void Run()
        {
            int templateCount = Templates.Count;
            if (templateCount > 0)
            {
                ClearErrMsgs();
                
                string[] templateArray = new string[templateCount];
                for (int i = 0; i < templateCount; i++)
                {
                    templateArray[i] = Templates[i].TemplateText;
                }
                Utility.WriteTrace("Compiling assembly...");
                CompilerResults results = CodeProvider.CompileAssemblyFromSource(CodeCompilerParameters, templateArray);
                Utility.WriteTrace("Done!");

                if (results.Errors.Count > 0 || results.CompiledAssembly == null)
                {
                    if (results.Errors.Count > 0)
                    {
                        foreach (CompilerError error in results.Errors)
                            LogErrMsgs("Compile Error: " + error.ErrorText);
                    }
                    if (results.CompiledAssembly == null)
                    {
                        string errorMessage = "Error generating template code: This usually indicates an error in template itself, such as use of reserved words. Detail: ";
                        Utility.WriteTrace(errorMessage + errMsg);
                        string sMessage = errorMessage + Environment.NewLine + errMsg;
                        throw new Exception(sMessage);
                    }
                    return;
                }

                Utility.WriteTrace("Extracting code from assembly and scrubbing output...");
                CallEntry(results.CompiledAssembly);
                Utility.WriteTrace("Done!");
            }
        }

        private static string ScrubOutput(string result)
        {
            if (!String.IsNullOrEmpty(result))
            {
                //the generator has an issue with adding extra lines. Trim them out
                
                result = regLineFix.Replace(result, "\r\n");

                //now, for readability, add a space after the end of the method/class
                result = Utility.FastReplace(result, "}", "}\r\n", StringComparison.InvariantCultureIgnoreCase);
                result = Utility.FastReplace(result, "namespace", "\r\nnamespace", StringComparison.InvariantCulture); //Must be case-sensitive, or it will cause VB End Namespace to wrap

                //these must remain as two separate procedures, or bad things will happen in VB
                result = Utility.FastReplace(result, "public class ", "\r\npublic class ", StringComparison.InvariantCulture); //trailing space need to address class names that begin with "class"
                result = Utility.FastReplace(result, "Public Class ", "\r\nPublic Class ", StringComparison.InvariantCulture); //trailing space need to address class names that begin with "Class"
                result = Utility.FastReplace(result, "[<]", "<", StringComparison.InvariantCultureIgnoreCase);
                result = Utility.FastReplace(result, "[>]", ">", StringComparison.InvariantCultureIgnoreCase);

                //This is should be executed last. While this value will ultimately be removed, it can be inserted in a template to keep an earlier operation from executing.
                //For example: <System.ComponentModel.DataObject()> Public Class MyController would normally wrap to a second line due upstream processing, which would
                //result in VB code that won't compile. However, <System.ComponentModel.DataObject()> Public [MONKEY_WRENCH]Class MyController, would not.

                //Nice Eric... :P
                //result = Utility.FastReplace(result, "[MONKEY_WRENCH]", String.Empty, StringComparison.InvariantCultureIgnoreCase);
                //Not currently used, but please leave in case we need it in the future!

            }
            return result;
        }


        internal StringCollection References = new StringCollection();

        //Flag that determines which languae to use
        internal ICodeLanguage Language = new CSharpCodeLanguage();

        private void CallEntry(Assembly assembly)
        {
            int templateCount = Templates.Count;
            Module mod = assembly.GetModules(false)[0];
            for (int i = 0; i < templateCount; i++)
            {
                Type type = mod.GetType(Templates[i].GeneratedRenderType);
                if (type != null)
                {
                    MethodInfo mi = type.GetMethod(Templates[i].EntryPoint, BindingFlags.Public | BindingFlags.NonPublic | BindingFlags.Static);
                    if (mi != null)
                    {
                        StringBuilder returnText = new StringBuilder();
                        
                        string resultMessage = "Successfully generated " + Templates[i].TemplateName;

                        if (Templates[i].AddUsingBlock)
                        {
                            Utility.WriteTrace("Adding Custom Using Blocks");
                            returnText.Append(Templates[i].RenderLanguage.DefaultUsingStatements);
                            returnText.Append(Templates[i].CustomUsingBlock);
                        }
                        try
                        {
                            returnText.Append((string)mi.Invoke(null, null));
                        }
                        catch (Exception ex)
                        {
                            resultMessage = "An Exception occured in template " + Templates[i].TemplateName + ": " + ex.Message;
                        }

                        Templates[i].FinalCode = ScrubOutput(returnText.ToString());
                        Utility.WriteTrace(resultMessage);
                    }
                }
            }
        }


        internal StringBuilder errMsg = new StringBuilder();
        internal void LogErrMsgs(string customMsg)
        {
            LogErrMsgs(customMsg, null);
        }

        internal void LogErrMsgs(string customMsg, Exception ex)
        {
            //put the error message into builder
            errMsg.Append("\r\n").Append(customMsg).Append(Environment.NewLine);

            //get all the exceptions and add their error messages
            while (ex != null)
            {
                errMsg.Append("\t").Append(ex.Message).Append(Environment.NewLine);
                ex = ex.InnerException;
            }
        }

        internal void ClearErrMsgs()
        {
            errMsg.Remove(0, errMsg.Length);
        }
    }

}
