using System;
using Microsoft.Practices.EnterpriseLibrary.Data;
using System.Data;
using System.Data.Common;

namespace SubSonic
{
    public class ELib3DataProvider : SqlDataProvider
    {
        #region Query Execution overrides
        public override IDataReader GetReader(QueryCommand qry)
        {
            Database db = DatabaseFactory.CreateDatabase();
            DbCommand cmd;
            if (qry.CommandType == CommandType.Text)
            {
                cmd = db.GetSqlStringCommand(qry.CommandSql);
            }
            else
            {
                cmd = db.GetStoredProcCommand(qry.CommandSql);

            }
            foreach (QueryParameter param in qry.Parameters)
            {
                db.AddInParameter(cmd, param.ParameterName, param.DataType, param.ParameterValue);
            }
            return db.ExecuteReader(cmd);

        }

		private static DbCommand GetDataSetCommand(Database db, QueryCommand qry)
		{
			DbCommand cmd;
			if (qry.CommandType == CommandType.Text)
				cmd = db.GetSqlStringCommand(qry.CommandSql);
			else
				cmd = db.GetStoredProcCommand(qry.CommandSql);

			foreach (QueryParameter param in qry.Parameters)
				db.AddInParameter(cmd, param.ParameterName, param.DataType, param.ParameterValue);

			return cmd;
		}

        public override T GetDataSet<T>(QueryCommand qry)
        {
			Database db = DatabaseFactory.CreateDatabase();
			DbCommand cmd = GetDataSetCommand(db, qry);

			T ds;
			if (typeof(T) == typeof(DataSet)) // LoadDataSet doesn't accept empty table names.
				ds = (T)db.ExecuteDataSet(cmd);
			else
			{
				ds = new T();
				db.LoadDataSet(cmd, ds, GetTableNames(ds));
			}
			return ds;
        }

        public override object ExecuteScalar(QueryCommand qry)
        {
            Database db = DatabaseFactory.CreateDatabase();
            DbCommand cmd;

            if (qry.CommandType == CommandType.Text)
            {
                cmd = db.GetSqlStringCommand(qry.CommandSql);
            }
            else
            {
                cmd = db.GetStoredProcCommand(qry.CommandSql);

            }
            cmd.CommandType = qry.CommandType;

            foreach (QueryParameter param in qry.Parameters)
            {
                db.AddInParameter(cmd, param.ParameterName, param.DataType, param.ParameterValue);
            }
            return db.ExecuteScalar(cmd);
        }
        public override int ExecuteQuery(QueryCommand qry)
        {
            Database db = DatabaseFactory.CreateDatabase();
            DbCommand cmd;

            if (qry.CommandType == CommandType.Text)
            {
                cmd = db.GetSqlStringCommand(qry.CommandSql);
            }
            else
            {
                cmd = db.GetStoredProcCommand(qry.CommandSql);

            }
            cmd.CommandType = qry.CommandType;

            foreach (QueryParameter param in qry.Parameters)
            {
                db.AddInParameter(cmd, param.ParameterName, param.DataType, param.ParameterValue);
            }
            return db.ExecuteNonQuery(cmd);
        }
        public override void ExecuteTransaction(QueryCommandCollection commands)
        {
            Database db = DatabaseFactory.CreateDatabase();

            //make sure we have at least one
            if (commands.Count > 0)
            {
                DbCommand cmd = null;

                //a using statement will make sure we close off the connection
                using (DbConnection conn = db.CreateConnection())
                {
                    //open up the connection and start the transaction
                    conn.Open();

                    DbTransaction trans = conn.BeginTransaction();

                    foreach (QueryCommand qry in commands)
                    {
                        if (qry.CommandType == CommandType.Text)
                        {
                            cmd = db.GetSqlStringCommand(qry.CommandSql);
                        }
                        else
                        {
                            cmd = db.GetStoredProcCommand(qry.CommandSql);
                        }

                        try
                        {
                            foreach (QueryParameter param in qry.Parameters)
                            {
                                db.AddInParameter(cmd, param.ParameterName, param.DataType, param.ParameterValue);
                            }

                            db.ExecuteNonQuery(cmd, trans);
                        }
                        catch (DbException x)
                        {
                            //if there's an error, roll everything back
                            trans.Rollback();

                            //clean up
                            conn.Close();
                            cmd.Dispose();

                            //throw the error retaining the stack.
                            throw new Exception(x.Message);
                        }
                    }
                    //if we get to this point, we're good to go
                    trans.Commit();

                    //close off the connection
                    conn.Close();
                    if (cmd != null)
                    {
                        cmd.Dispose();
                    }
                }
            }
            else
            {
                throw new Exception("No commands present");
            }
        }

        #endregion

        #region Command Builder Overrides
        public override DbCommand GetDbCommand(QueryCommand qry)
        {
            Database db = DatabaseFactory.CreateDatabase();
            DbCommand cmd;
            if (qry.CommandType == CommandType.StoredProcedure)
            {
                cmd = db.GetStoredProcCommand(qry.CommandSql);
            }
            else
            {
                cmd = db.GetSqlStringCommand(qry.CommandSql);
            }

            foreach (QueryParameter par in qry.Parameters)
            {
                db.AddInParameter(cmd, par.ParameterName, par.DataType, par.ParameterValue);
            }

            return cmd;
        }
        public override IDbCommand GetCommand(QueryCommand qry)
        {

            Database db = DatabaseFactory.CreateDatabase();
            DbCommand cmd = db.GetSqlStringCommand(qry.CommandSql);
            foreach (QueryParameter param in qry.Parameters)
            {
                db.AddInParameter(cmd, param.ParameterName, param.DataType, param.ParameterValue);
            }

            return cmd;
        }
        #endregion


    }

    /// <summary>
    /// This is class is for backwards compatibility only. Please update your references to ELib3DataProvider
    /// </summary>
    public class ELib2DataProvider : ELib3DataProvider
    {
    }
}
