''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Code author: Martin Lapierre, http://devinstinct.com
''''''''''''''''''''''''''''''''''''''''''''''''''''''

Imports System
Imports EnvDTE
Imports EnvDTE80

Namespace SubSonic.Tools

    ''' <summary>
    ''' Handles DTE events to keep the files synchronized with SubSonic files.
    ''' </summary>
    ''' <remarks>
    ''' Class not thread-safe. 
    ''' Exceptions catched while handling events are displayed in message boxes.
    ''' </remarks>
    Public Class EnvDTEEventManager


#Region "Constructors"

        ''' <summary>
        ''' Constructor.
        ''' Makes this Event Manager handle the events of the given DTE.
        ''' </summary>
        ''' <param name="DTE">DTE to handle the events for.</param>
        Public Sub New(ByVal DTE As DTE2)
            miscFilesEvents(1) = RegisterFileEvents(DTE, "CSharpProjectItemsEvents")
            miscFilesEvents(2) = RegisterFileEvents(DTE, "VBProjectItemsEvents")
            documentEvents = CType(DTE.Events.DocumentEvents(Nothing), EnvDTE.DocumentEvents)
        End Sub

#End Region


#Region "Methods"

        ''' <summary>
        ''' Returns the registered file events of the given DTE and project items.
        ''' </summary>
        ''' <param name="DTE">DTE to register the events for.</param>
        ''' <param name="objectName">Name of the project items events object.</param>
        ''' <returns>The registered file events.</returns>
        Private Function RegisterFileEvents(ByVal DTE As DTE2, ByVal objectName As String) As EnvDTE.ProjectItemsEvents
            Dim miscFilesEvents As EnvDTE.ProjectItemsEvents
            miscFilesEvents = CType(DTE.Events.GetObject(objectName), EnvDTE.ProjectItemsEvents)
            AddHandler miscFilesEvents.ItemRenamed, AddressOf miscFilesEventsBoth_ItemRenamed
            Return miscFilesEvents
        End Function

#End Region


#Region "Event Handlers"

        ''' <summary>
        ''' Occurs immediately after a project is renamed in a solution or a project item is renamed in a project. 
        ''' </summary>
        ''' <param name="ProjectItem">The renamed project item.</param>
        ''' <param name="OldName">The item's old name.</param>
        Private Sub miscFilesEventsBoth_ItemRenamed(ByVal projectItem As EnvDTE.ProjectItem, ByVal oldName As String)
            Try
                ' Handle only if the name has been changed 
                ' (sometimes the event is received for sub-items whose names were not changed).
                If (projectItem.Name <> oldName) Then
                    ' Handle rename only if the item already contains an enum resource file.
                    Dim relativeCount As Integer = SubSonicEnvDTE.GetDALFileRelativeCount(projectItem, oldName)
                    If (relativeCount > 0) Then
                        SubSonicEnvDTE.Rename(projectItem, oldName)
                    End If
                End If
            Catch ex As Exception
                MsgBox(ex.Message)
            End Try
        End Sub

        ''' <summary>
        ''' Occurs when a document is saved. 
        ''' </summary>
        ''' <param name="Document">The saved document.</param>
        Private Sub documentEvents_DocumentSaved(ByVal document As EnvDTE.Document) Handles documentEvents.DocumentSaved
            Try
                ' Generate only if the document already contains generated file(s).
                Dim relativeCount As Integer = SubSonicEnvDTE.GetDALFileRelativeCount(document)
                If (relativeCount = 1) Then
                    SubSonicEnvDTE.Create(document, GenerationKinds.SingleFile)
                ElseIf (relativeCount > 1) Then
                    SubSonicEnvDTE.Create(document, GenerationKinds.MultipleFiles)
                End If
            Catch ex As Exception
                MsgBox(ex.Message)
            End Try
        End Sub

#End Region


#Region "Fields"

        ''' <summary>
        ''' Array of files events.
        ''' </summary>
        Private miscFilesEvents(3) As EnvDTE.ProjectItemsEvents

        ''' <summary>
        ''' Document events.
        ''' </summary>
        <System.ContextStaticAttribute()> Private WithEvents documentEvents As EnvDTE.DocumentEvents

#End Region


    End Class

End Namespace
