''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Code author: Martin Lapierre, http://devinstinct.com
''''''''''''''''''''''''''''''''''''''''''''''''''''''

Imports System.CodeDom
Imports System.Collections.Generic
Imports System.IO
Imports System.Text
Imports System.Windows.Forms
Imports EnvDTE
Imports SubSonic

Namespace SubSonic.Tools

    Public Module SubSonicEnvDTE

        Sub New()
            Application.EnableVisualStyles()
        End Sub

        ''' <summary>
        ''' Indicates if the given project item is a SubSonic file.
        ''' </summary>
        ''' <param name="itemInfo">The item information to check.</param>
        ''' <returns>true if the given item is a valid SubSonic .abp file; false otherwise.</returns>
        Public Function IsSubSonicFile(ByVal itemInfo As ItemInfo) As Boolean
            ' Validate parameters
            If (itemInfo Is Nothing) Then
                Throw New ArgumentNullException("itemInfo")
            End If
            If (itemInfo.Kind = EnvDTE.Constants.vsProjectItemKindPhysicalFile And Path.GetExtension(itemInfo.FileName) = ".abp") Then
                Return True
            Else
                Return False
            End If
        End Function

        Public Function IsSubSonicFile(ByVal document As Document) As Boolean
            ' Validate parameters
            If (document Is Nothing) Then
                Throw New ArgumentNullException("document")
            End If
            Dim itemInfo As ItemInfo
            If (document.ProjectItem IsNot Nothing) Then
                itemInfo = New ProjectItemInfo(document.ProjectItem)
                Return IsSubSonicFile(itemInfo)
            Else
                Return False
            End If
        End Function

        Public Function GetDALFileRelativeCount(ByVal document As Document) As Integer
            If (document Is Nothing) Then
                Throw New ArgumentNullException("document")
            End If
            If (IsSubSonicFile(document)) Then
                Return GetDALFileRelativeCount(document.ProjectItem, Nothing)
            Else
                Return 0
            End If
        End Function

        Public Function GetDALFileRelativeCount(ByVal projectItem As ProjectItem, ByVal oldName As String) As Integer
            If (projectItem Is Nothing) Then
                Throw New ArgumentNullException("projectItem")
            End If
            Dim itemInfo As ItemInfo
            itemInfo = New ProjectItemInfo(projectItem)
            If (IsSubSonicFile(itemInfo)) Then
                Return projectItem.ProjectItems.Count
            ElseIf (projectItem.Kind = EnvDTE.Constants.vsProjectItemKindPhysicalFolder Or projectItem.ProjectItems Is Nothing) Then
                Dim collection As ProjectItems = Nothing
                If (projectItem.ProjectItems Is Nothing) Then ' Web Application project
                    collection = projectItem.Collection
                Else
                    collection = projectItem.ProjectItems
                End If
                Dim relativeCount As Integer = 0
                Dim targetPrefix As String = Nothing
                If (String.IsNullOrEmpty(oldName)) Then
                    targetPrefix = Path.GetFileNameWithoutExtension(projectItem.Name)
                Else
                    targetPrefix = Path.GetFileNameWithoutExtension(oldName)
                End If
                Dim extension As String = GetProjectFileExtension(projectItem.ContainingProject)
                For Each subItem As ProjectItem In collection
                    If subItem.Name.StartsWith(targetPrefix & ".") And subItem.Name.EndsWith(extension) Then
                        If subItem.Name = targetPrefix & extension Then
                            relativeCount = 1 ' Assume single file
                        Else
                            relativeCount = 2 ' Assume multiple file (value of 2 for more than 1 file)
                        End If
                        Exit For
                    End If
                Next
                Return relativeCount
            Else
                Return 0
            End If
        End Function

        Public Function GetDALFileItems(ByVal projectItem As ProjectItem, ByVal oldName As String) As List(Of ProjectItem)
            If (projectItem Is Nothing) Then
                Throw New ArgumentNullException("projectItem")
            End If
            Dim itemInfo As ItemInfo
            itemInfo = New ProjectItemInfo(projectItem)
            Dim dalItems As New List(Of ProjectItem)
            If (IsSubSonicFile(itemInfo)) Then
                For Each subItem As ProjectItem In projectItem.ProjectItems
                    dalItems.Add(subItem)
                Next
            ElseIf (projectItem.Kind = EnvDTE.Constants.vsProjectItemKindPhysicalFolder Or projectItem.ProjectItems Is Nothing) Then
                Dim collection As ProjectItems = Nothing
                If (projectItem.ProjectItems Is Nothing) Then ' Web Application project
                    collection = projectItem.Collection
                Else
                    collection = projectItem.ProjectItems
                End If
                Dim relativeCount As Integer = 0
                Dim targetPrefix As String = Nothing
                If (String.IsNullOrEmpty(oldName)) Then
                    targetPrefix = Path.GetFileNameWithoutExtension(projectItem.Name)
                Else
                    targetPrefix = Path.GetFileNameWithoutExtension(oldName)
                End If
                Dim extension As String = GetProjectFileExtension(projectItem.ContainingProject)
                For Each subItem As ProjectItem In collection
                    If subItem.Name.StartsWith(targetPrefix & ".") And subItem.Name.EndsWith(extension) Then
                        dalItems.Add(subItem)
                    End If
                Next
            End If
            Return dalItems
        End Function

        Public Function Rename(ByVal projectItem As ProjectItem, ByVal oldName As String) As Integer
            If (projectItem Is Nothing) Then
                Throw New ArgumentNullException("projectItem")
            End If
            Dim dalItems As List(Of ProjectItem) = GetDALFileItems(projectItem, oldName)

            ' Delete obsolete generated files.
            If (dalItems.Count > 0) Then
                Dim fileCount As Integer = 0
                Dim filesToRename As New StringBuilder()
                For Each item As ProjectItem In dalItems
                    fileCount += 1
                    If (fileCount < 20) Then
                        If (filesToRename.Length > 0) Then
                            filesToRename.Append(", ")
                        End If
                        filesToRename.Append(item.Name)
                    Else
                        filesToRename.Append("... (more files)")
                        Exit For
                    End If
                Next
                Dim targetPrefix As String = Path.GetFileNameWithoutExtension(projectItem.Name)
                Dim oldPrefix As String = Path.GetFileNameWithoutExtension(oldName)
                Select Case MessageBox.Show(String.Format("SubSonic has detected that the following file(s) should probably be renamed from '{2}...' to '{3}...':{0}{0}{1}{0}{0}Do you wish to proceed with renaming?", ControlChars.CrLf, filesToRename, oldPrefix, targetPrefix), "SubSonic", MessageBoxButtons.YesNo, MessageBoxIcon.Question)
                    Case DialogResult.Yes
                        For Each item As ProjectItem In dalItems
                            item.Name = targetPrefix & item.Name.Remove(0, oldPrefix.Length)
                        Next
                End Select
            End If
        End Function

        ''' <summary>
        ''' Gets the clanguage type of the given project item code model.
        ''' </summary>
        ''' <param name="project">The project to get the language type for.</param>
        ''' <returns>The language type.</returns>
        Private Function GetProjectLanguageType(ByVal project As Project) As LanguageType
            Select Case Project.CodeModel.Language
                Case CodeModelLanguageConstants.vsCMLanguageCSharp
                    Return LanguageType.CSharp
                Case CodeModelLanguageConstants.vsCMLanguageVB
                    Return LanguageType.VB
                Case Else
                    Throw New Exception("Unsupported programming language.")
            End Select
        End Function

        ''' <summary>
        ''' Gets the code file extension of the given project item code model.
        ''' </summary>
        ''' <param name="project">The project to get the code file extension for.</param>
        ''' <returns>The code file extension.</returns>
        Private Function GetProjectFileExtension(ByVal project As Project) As String
            Select Case project.CodeModel.Language
                Case CodeModelLanguageConstants.vsCMLanguageCSharp
                    Return FileExtension.DOT_CS
                Case CodeModelLanguageConstants.vsCMLanguageVB
                    Return FileExtension.DOT_VB
                Case Else
                    Throw New Exception("Unsupported programming language.")
            End Select
        End Function

		Public Sub Create(ByVal document As Document, ByVal generationKind As GenerationKinds)
			Try
				If (IsSubSonicFile(document)) Then
					Dim itemInfo As ItemInfo
					itemInfo = New ProjectItemInfo(document.ProjectItem)
					Create(itemInfo, generationKind)
				Else
					Throw New Exception(String.Format("Can't generate single file from {0}; the target must be a .abp file.", document.Name))
				End If
			Catch ex As Exception
				HandleException(ex)
			End Try
		End Sub

        ''' <summary>
        ''' Creates the DAL files for the given selected items.
        ''' </summary>
        ''' <param name="selectedItems">The items to create the DAL files for.</param>
        ''' <remarks>
        ''' On error, the Exception Data contains the Exceptions that occured for specific items.
        ''' For each Data entry, the Key is the SelectedItem name and index 
        ''' (in the form Name:Index) and the Value is the Exception. 
        ''' </remarks>
		Public Sub Create(ByVal selectedItems As SelectedItems, ByVal generationKind As GenerationKinds)
			Try
				' Validate parameters
				If (selectedItems Is Nothing) Then
					Throw New ArgumentNullException("selectedItems")
				End If
				Dim itemIndex As Integer
				Dim exception As New Exception("Can't generate DAL for some files.")
				For Each selectedItem As SelectedItem In selectedItems
					Try
						'Solution
						'   Project = Nothing
						'   ProjectItems = Nothing
						'Project
						'   Project OK
						'   ProjectItems = Nothing
						'Folder
						'   Project = Nothing
						'   ProjectItems OK -> Kind = EnvDTE.Constants.vsProjectItemKindPhysicalFolder
						'File
						'   Project = Nothing
						'   ProjectItems OK -> Kind = EnvDTE.Constants.vsProjectItemKindPhysicalFile
						Dim itemInfo As ItemInfo
						If (selectedItem.Project IsNot Nothing) Then
							itemInfo = New ProjectInfo(selectedItem.Project)
						ElseIf (selectedItem.ProjectItem IsNot Nothing) Then
							itemInfo = New ProjectItemInfo(selectedItem.ProjectItem)
						Else
							Throw New Exception(String.Format("Can't generate DAL files from {0}", selectedItem.Name))
						End If
						Create(itemInfo, generationKind)
					Catch ex As Exception
						If (Not exception.Data.Contains(selectedItem.Name)) Then
							exception.Data.Add(String.Format("{0}:{1}", selectedItem.Name, itemIndex), ex)
						End If
					Finally
						itemIndex += 1
					End Try
				Next
				If (exception.Data.Count > 0) Then
					Throw exception
				End If
			Catch ex As Exception
				HandleException(ex)
			End Try
		End Sub

		Public Sub Create(ByVal itemInfo As ItemInfo, ByVal generationKind As GenerationKinds)
			Try
				Select Case generationKind
					Case GenerationKinds.FromSelection
						GenerateFromSelection(itemInfo)
					Case GenerationKinds.MultipleFiles
						GenerateMultipleFiles(itemInfo)
					Case GenerationKinds.SingleFile
						GenerateSingleFile(itemInfo)
				End Select
			Catch ex As Exception
				HandleException(ex)
			End Try
		End Sub

        Private Sub GenerateFromSelection(ByVal itemInfo As ItemInfo)
			If (IsSubSonicFile(itemInfo) Or itemInfo.Kind = EnvDTE.Constants.vsProjectItemKindPhysicalFolder Or TypeOf (itemInfo) Is ProjectInfo) Then
				Dim host As New SubSonicHost(itemInfo.FileName)
				Dim genForm As New GenerationForm(host)
				Dim compileUnits As Dictionary(Of String, CodeCompileUnit) = Nothing

				genForm.ShowDialog()
				If (Not genForm.MustGenerate) Then
					Throw New Exception("Code generation cancelled.")
				End If

				For Each dataEntry As KeyValuePair(Of String, ProviderData) In genForm.ProviderData
					If dataEntry.Value.MustGenerate Then
						Dim previousCursor As Cursor
						previousCursor = Cursor.Current
						Cursor.Current = Cursors.WaitCursor
						Try
							compileUnits = host.SubSonic.GenerateMultipleUnits(GetProjectLanguageType(itemInfo.Project), dataEntry.Value.SelectedTables, dataEntry.Value.SelectedViews, dataEntry.Value.SelectedSPs, dataEntry.Value.AddDependencies, dataEntry.Key)
							GenerateDAL(itemInfo, compileUnits, dataEntry.Key)
						Finally
							Cursor.Current = previousCursor
						End Try
					End If
				Next
			Else
				Throw New Exception(String.Format("Can't generate from selection from {0}; the target must be a folder, a project or an .abp file.", itemInfo.Name))
			End If
        End Sub

        Private Sub GenerateMultipleFiles(ByVal itemInfo As ItemInfo)
			If (IsSubSonicFile(itemInfo) Or itemInfo.Kind = EnvDTE.Constants.vsProjectItemKindPhysicalFolder Or TypeOf (itemInfo) Is ProjectInfo) Then
				Dim previousCursor As Cursor
				previousCursor = Cursor.Current
				Cursor.Current = Cursors.WaitCursor
				Try
					Dim abpFileName As String = itemInfo.FileName
					Dim host As New SubSonicHost(abpFileName)
					Dim compileUnits As Dictionary(Of String, CodeCompileUnit) = Nothing
					compileUnits = host.SubSonic.GenerateMultipleUnits(GetProjectLanguageType(itemInfo.Project))
					GenerateDAL(itemInfo, compileUnits, Nothing)
				Finally
					Cursor.Current = previousCursor
				End Try
			Else
				Throw New Exception(String.Format("Can't generate multiple files from {0}; the target must be a folder, a project or an .abp file.", itemInfo.Name))
			End If
        End Sub

        Private Sub GenerateSingleFile(ByVal itemInfo As ItemInfo)
			If (IsSubSonicFile(itemInfo) Or itemInfo.Kind = EnvDTE.Constants.vsProjectItemKindPhysicalFolder Or TypeOf (itemInfo) Is ProjectInfo) Then
				Dim previousCursor As Cursor
				previousCursor = Cursor.Current
				Cursor.Current = Cursors.WaitCursor
				Try
					Dim abpFileName As String = itemInfo.FileName
					Dim host As New SubSonicHost(abpFileName)
					Dim compileUnits As New Dictionary(Of String, CodeCompileUnit)
					Dim unit As CodeCompileUnit = host.SubSonic.GenerateSingleUnit(GetProjectLanguageType(itemInfo.Project))
					compileUnits.Add(String.Empty, unit)
					GenerateDAL(itemInfo, compileUnits, Nothing)
				Finally
					Cursor.Current = previousCursor
				End Try
			Else
				Throw New Exception(String.Format("Can't generate single file from {0}; the target must be a folder, a project or an .abp file.", itemInfo.Name))
			End If
        End Sub

        Private Function CanWriteFile(ByVal fileName As String) As Boolean

            If (Not File.Exists(fileName)) Then
                Return True
            End If

            While True
                If ((File.GetAttributes(fileName) And FileAttributes.ReadOnly) = FileAttributes.ReadOnly) Then
                    Select Case MessageBox.Show(String.Format("The file '{0}' is read-only and cannot be modified.", fileName), "SubSonic", MessageBoxButtons.AbortRetryIgnore, MessageBoxIcon.Warning)
                        Case DialogResult.Abort
                            Throw New Exception("Code generation aborted by user.")
                        Case DialogResult.Retry
                            Continue While
                        Case Else ' Ignore
                            Return False
                    End Select
                Else
                    Return True
                End If
            End While

        End Function

        ''' <summary>
        ''' Launches the code generation of the SubSonic file for the given resx project item.
        ''' </summary>
        ''' <param name="itemInfo">The item information from which to create the DAL files for.</param>
        ''' <remarks></remarks>
        Private Sub GenerateDAL(ByVal itemInfo As ItemInfo, ByVal compileUnits As Dictionary(Of String, CodeCompileUnit), ByVal providerName As String)

            Dim extension As String = GetProjectFileExtension(itemInfo.Project)
            Dim currentFiles As New Dictionary(Of String, ProjectItem)
            Dim targetPrefix As String = Path.GetFileNameWithoutExtension(itemInfo.Name) & "." & providerName ' providerName can be Nothing

            ' Store a list of the current files.
            For Each subItem As ProjectItem In itemInfo.RelatedProjectItems
                If (String.Compare(subItem.Name, itemInfo.Name, True) <> 0) Then
                    If (subItem.Name.StartsWith(targetPrefix, StringComparison.CurrentCultureIgnoreCase) And subItem.Name.EndsWith(extension, StringComparison.CurrentCultureIgnoreCase)) Then
                        currentFiles.Add(subItem.FileNames(1).ToLower(), subItem)
                    End If
                End If
            Next

            ' Generate new / update existing files.
            For Each unitEntry As KeyValuePair(Of String, CodeCompileUnit) In compileUnits
                Dim genFileName As String = Nothing
                If unitEntry.Key = String.Empty Then
                    genFileName = Path.GetDirectoryName(itemInfo.FileName) & Path.DirectorySeparatorChar & Path.GetFileNameWithoutExtension(itemInfo.Name) & extension
                Else
                    genFileName = Path.GetDirectoryName(itemInfo.FileName) & Path.DirectorySeparatorChar & Path.GetFileNameWithoutExtension(itemInfo.Name) & "." & unitEntry.Key
                End If
                If (CanWriteFile(genFileName)) Then
                    Using fileStream As New StreamWriter(genFileName)
                        If TypeOf (unitEntry.Value) Is CodeSnippetCompileUnit Then
                            fileStream.WriteLine(DirectCast(unitEntry.Value, CodeSnippetCompileUnit).Value)
                        Else
                            '' TODO: non-CodeSnippet support
                        End If
                    End Using

                    ' Add file to project.
                    If (Not currentFiles.ContainsKey(genFileName.ToLower())) Then
                        itemInfo.RelatedProjectItems.AddFromFile(genFileName)
                    End If
                End If

                ' Remove from current file list. 
                ' Those who will remain in the list at the end of the loop 
                ' will be the files that must be deleted (they haven't been generated this time).
                If (currentFiles.ContainsKey(genFileName.ToLower())) Then
                    currentFiles.Remove(genFileName.ToLower())
                End If
            Next

            ' Delete obsolete generated files.
            If (currentFiles.Count > 0) Then
                Dim fileCount As Integer = 0
                Dim fileToDelete As New StringBuilder()
                For Each oldProjectItem As KeyValuePair(Of String, ProjectItem) In currentFiles
                    fileCount += 1
                    If (fileCount < 20) Then
                        If (fileToDelete.Length > 0) Then
                            fileToDelete.Append(", ")
                        End If
                        fileToDelete.Append(oldProjectItem.Value.Name)
                    Else
                        fileToDelete.Append("... (more files)")
                        Exit For
                    End If
                Next
				Select Case MessageBox.Show(String.Format("SubSonic has detected that the following file(s) should probably be deleted:{0}{0}{1}{0}{0}Do you wish to proceed with deletion (this may take a while)?", ControlChars.CrLf, fileToDelete), "SubSonic", MessageBoxButtons.YesNo, MessageBoxIcon.Question)
					Case DialogResult.Yes
						For Each oldProjectItem As KeyValuePair(Of String, ProjectItem) In currentFiles
							oldProjectItem.Value.Delete()
						Next
				End Select
            End If

        End Sub

		''' <summary>Handles a given exception.</summary>
		Private Sub HandleException(ByVal ex As Exception)
			If (ex.Data.Count = 0) Then	' Standard exception.
				DisplayMessage(ex)
			ElseIf (ex.Data.Count = 1) Then	' Exception collection with one entry.
				For Each entry As DictionaryEntry In ex.Data
					Dim targetException As Exception = CType(entry.Value, Exception)
					DisplayMessage(targetException)
				Next
			Else ' Exception collection with many entries.
				Dim files As String = Nothing
				Dim itemException As Exception
				For Each entry As DictionaryEntry In ex.Data
					itemException = CType(entry.Value, Exception)
					If (files = Nothing) Then
						files = String.Format("{0}[{1}]", entry.Key.ToString(), itemException.Message)
					Else
						files = String.Format("{0}, {1}[{2}]", files, entry.Key.ToString(), itemException.Message)
					End If
				Next
				MessageBox.Show(String.Format("{0} ({1})", ex.Message, files), "SubSonic", MessageBoxButtons.OK, MessageBoxIcon.Information)
			End If
		End Sub

		''' <summary>Displays the message of the given exception.</summary>
		Public Sub DisplayMessage(ByVal ex As Exception)
			If (TypeOf ex Is System.Reflection.TargetInvocationException And ex.InnerException IsNot Nothing) Then
				MessageBox.Show(ex.InnerException.Message, "SubSonic", MessageBoxButtons.OK, MessageBoxIcon.Information)
			ElseIf (TypeOf ex Is System.Runtime.InteropServices.COMException) Then
				MessageBox.Show(String.Format("The process may have timed out, please retry: {0}", ex.Message), "SubSonic", MessageBoxButtons.OK, MessageBoxIcon.Information)
			Else
				MessageBox.Show(ex.Message, "SubSonic", MessageBoxButtons.OK, MessageBoxIcon.Information)
			End If
		End Sub
    End Module

End Namespace
