///////////////////////////////////////////////////////
// Code author: Martin Lapierre, http://devinstinct.com
///////////////////////////////////////////////////////

using System;
using SubSonic.CodeGenerator;
using SubSonic.Utilities;
using System.Text;
using System.Collections.Generic;
using System.CodeDom;
using System.Collections.Specialized;

namespace SubSonic.Tools
{
	public static class BuildService
	{
		// Mostly code from SubSonic::BuildProvider::BuildCode()
		public static string GenerateCode(ICodeLanguage language)
		{
			//make sure the providers are loaded
			Utility.WriteTrace("Initializing Providers....");
			DataService.LoadProviders();

			StringBuilder code = new StringBuilder(language.DefaultUsingStatements);

			TurboCompiler tc = new TurboCompiler();
			//loop over the providers, and generate code for each
			foreach (DataProvider provider in DataService.Providers)
			{
				Utility.WriteTrace("Creating code for " + provider.Name);
				string[] tableList = DataService.GetTableNames(provider.Name);
				string[] viewList = DataService.GetViewNames(provider.Name);


				foreach (string tbl in tableList)
				{
					TurboTemplate tt = CodeService.BuildClassTemplate(tbl, language, provider);
					tc.AddTemplate(tt);
					TurboTemplate tODS = CodeService.BuildODSTemplate(tbl, language, provider);
					tc.AddTemplate(tODS);
				}

				foreach (string view in viewList)
				{
					TurboTemplate tt = CodeService.BuildViewTemplate(view, language, provider);
					tc.AddTemplate(tt);
				}

				tc.AddTemplate(CodeService.BuildSPTemplate(language, provider));
			}

			if (DataService.Providers.Count > 0)
			{
				tc.AddTemplate(CodeService.BuildStructsTemplate(language, DataService.Provider));
			}

			foreach (TurboTemplate tt in tc.Templates)
			{
				tt.AddUsingBlock = false;
			}
			tc.Run();
			foreach (TurboTemplate tt in tc.Templates)
			{
				code.Append(tt.FinalCode);
			}
			return code.ToString();
		}

		/// <summary>
		/// Generates the DAL in a single CodeCompileUnit from a table list.
		/// </summary>
		/// <param name="languageType">The language to generate the DAL for.</param>
		/// <param name="tables">The table names, one per line in the string.</param>
		/// <returns>A single CodeCompileUnit.</returns>
		public static CodeCompileUnit GenerateSingleUnit(ICodeLanguage language)
		{
			string code = GenerateCode(language);

			return new CodeSnippetCompileUnit(code);
		}

		/// <summary>
		/// Generates the DAL in multiple CodeCompileUnits for specific tables, views
		/// and, optionnaly, all stored procedures.
		/// </summary>
		/// <param name="languageType">The language to generate the DAL for.</param>
		/// <param name="tables">
		/// The list of tables to generate the DAL for; first element as "*" for all tables.
		/// Can be null or empty. A null value generates no table and no table struct. 
		/// <param name="views">
		/// The list of views to generate the DAL for; first element as "*" for all views.
		/// Can be null or empty. A null value generates no view and no view struct. 
		/// </param>
		/// <param name="sps">
		/// The list of stored procedures to generate the DAL for; first element as "*" for all stored procedures.
		/// Can be null or empty. A null value generates no stored procedures. 
		/// </param>
		/// <param name="providerName">
		/// The name of the provider to generate the DAL for.
		/// If null, generates code for all the providers.
		/// </param>
		/// <param name="addDependencies">Indicates whether to add the dependent tables or not.</param>
		/// <returns>A dictionary of suggested filenames/CodeCompileUnit pairs.</returns>
		public static Dictionary<string, CodeCompileUnit> GenerateMultipleUnits(ICodeLanguage language, string[] tables, string[] views, string[] sps, bool addDependencies, string providerName)
		{
			string header = null;
			string keyName = null;
			List<string> providerNames = new List<string>();
			Dictionary<string, CodeCompileUnit> units = new Dictionary<string, CodeCompileUnit>();

			// Init language-dependant variables.
			header = ParseUtility.GetUsings(language);

			// Make sure the providers are loaded.
			DataService.LoadProviders();

			// Build the list of the providers to generate the code for.
			if (providerName != null)
				providerNames.Add(providerName);
			else // If none provided, generate for all.
				foreach (DataProvider provider in DataService.Providers)
					providerNames.Add(provider.Name);

			TurboCompiler turboCompiler = new TurboCompiler();
			TurboTemplate tt = null;

			// Generate code for all selected providers.
			foreach (string loopProviderName in providerNames)
			{
				DataProvider provider = DataService.GetInstance(loopProviderName);

				// Override config for tables & views.
				provider.IncludeTables = new string[] { "*" };
				provider.ExcludeTables = new string[0];

				// Generate Tables & Table struct.
				if (tables != null)
				{
					string[] tablesToGenerate = null;
					if (tables.Length == 1 && tables[0].Trim() == "*")
						tablesToGenerate = DataService.GetTableNames(provider.Name);
					else
						tablesToGenerate = tables;

					List<string> dependencyTables = new List<string>();

					// Tables
					// Must be called first; GenerateTableStruct dependent on table list.
					Dictionary<string, string> generatedTables = new Dictionary<string, string>(); // Keeps track of generated tables (for dependencies generation).
					foreach (string table in tablesToGenerate)
						AddTableAndDependencies(turboCompiler, generatedTables, table, provider, language, addDependencies);

					// ODS Controllers
					// Must be called first; GenerateTableStruct dependent on table list.
					foreach (string table in tablesToGenerate)
					{
						keyName = DataService.GetSchema(table, provider.Name, TableType.Table).ClassName + "Controller";
						tt = CodeService.BuildODSTemplate(table, language, provider);
						tt.OutputPath = GetProviderFileName(provider, keyName, language);
						turboCompiler.AddTemplate(tt);
					}
				}

				// Generate Views & View struct.
				if (views != null)
				{
					string[] viewsToGenerate = null;
					if (views.Length == 1 && views[0].Trim() == "*")
						viewsToGenerate = DataService.GetViewNames(provider.Name);
					else
						viewsToGenerate = views;

					// Views
					foreach (string view in viewsToGenerate)
					{
						keyName = DataService.GetSchema(view, provider.Name, TableType.View).ClassName;
						tt = CodeService.BuildViewTemplate(view, language, provider);
						tt.OutputPath = GetProviderFileName(provider, keyName, language);
						turboCompiler.AddTemplate(tt);
					}
				}

				// Generate Stored Procedures.
				if (sps != null)
				{
					// SPs must be explicitely specified in the provider.
					provider.IncludeProcedures = sps;
					provider.ExcludeProcedures = new string[0];
					//codeString = CodeService.RunSPs(provider.Name, language);
					try
					{
						keyName = provider.SPClassName.Replace(" ", "").Trim();
					}
					catch
					{
						keyName = ClassName.STORED_PROCEDURES;
					}

					tt = CodeService.BuildSPTemplate(language, provider);
					tt.OutputPath = GetProviderFileName(provider, keyName, language);
					turboCompiler.AddTemplate(tt);
				}
			}

			// Generate structs/database for all providers.
			keyName = "Structures";
			tt = CodeService.BuildStructsTemplate(language, DataService.Provider);
			tt.OutputPath = keyName + language.FileExtension;
			turboCompiler.AddTemplate(tt);

			turboCompiler.Run();

			// Generate compile units from compiler.
			foreach (TurboTemplate template in turboCompiler.Templates)
				units.Add(template.OutputPath, new CodeSnippetCompileUnit(template.FinalCode));

			return units;
		}

		public static string GetProviderFileName(DataProvider provider, string keyName, ICodeLanguage language)
		{
			return Utility.GetParameterName(provider.Name, provider) + "." + keyName + language.FileExtension;
		}

		/// <summary>
		/// Adds the generated table class to the given codeData collection  
		/// and recursively adds dependant tables as well (used to handle many-to-many, pk & fk).
		/// </summary>
		/// <param name="generatedTables">The dictionary to which to add the generated classes.</param>
		/// <param name="table">The table to add the code for.</param>
		/// <param name="provider">The data provider to use.</param>
		/// <param name="languageType">The language to generate the code for.</param>
		/// <param name="addDependencies">Indicates whether to add the dependent tables or not.</param>
		private static void AddTableAndDependencies(TurboCompiler turboCompiler, Dictionary<string, string> generatedTables, string table, DataProvider provider, ICodeLanguage language, bool addDependencies)
		{
			string keyName = DataService.GetSchema(table, provider.Name, TableType.Table).ClassName;
			string fileName = GetProviderFileName(provider, keyName, language);
			if (!generatedTables.ContainsKey(fileName)) // Add only once; use filename since multiple providers may have same table names.
			{
				TurboTemplate tt = CodeService.BuildClassTemplate(table, language, provider);
				tt.OutputPath = fileName;
				turboCompiler.AddTemplate(tt);
				generatedTables.Add(tt.OutputPath, table);

				if (addDependencies)
				{
					TableSchema.Table schemaTable = DataService.GetSchema(table, provider.Name, TableType.Table);
					if (schemaTable.ManyToManys != null)
						foreach (TableSchema.ManyToManyRelationship relationship in schemaTable.ManyToManys)
							AddTableAndDependencies(turboCompiler, generatedTables, relationship.ForeignTableName, provider, language, addDependencies);

					if (schemaTable.PrimaryKeyTables != null)
						foreach (TableSchema.PrimaryKeyTable pkTable in schemaTable.PrimaryKeyTables)
							AddTableAndDependencies(turboCompiler, generatedTables, pkTable.TableName, provider, language, addDependencies);

					if (schemaTable.ForeignKeys != null)
						foreach (TableSchema.ForeignKeyTable fkTable in schemaTable.ForeignKeys)
							AddTableAndDependencies(turboCompiler, generatedTables, fkTable.TableName, provider, language, addDependencies);
				}
			}
		}
	}
}
