///////////////////////////////////////////////////////
// Code author: Martin Lapierre, http://devinstinct.com
///////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.IO;
using System.Reflection;
using System.Runtime.CompilerServices;

namespace SubSonic.Tools
{
	/// <summary>
	/// Hosts SubSonic for any environment.
	/// </summary>
	/// <remarks>
	/// This host allows multiple SubSonic code generators 
	/// to run simultaneously even if the SubSonic core classes are static
	/// and also makes SubSonic work with Enterprise Library from an
	/// out-of-process tool. 
	/// This is done by spawning a dedicated AppDomain. 
	/// </remarks>
	public class SubSonicHost
	{
		#region Fields

		/// <summary>
		/// The type of the internal generator.
		/// </summary>
		/// <remarks>
		/// More reliable than using a hard-coded string.
		/// </remarks>
		private static Type _generatorType = typeof(SubSonicProxy);

		/// <summary>
		/// The hosted compile unit generator.
		/// </summary>
		private ISubSonic _subsonic = null;

		#endregion Fields


		#region Constructors

		/// <summary>
		/// Constructor.
		/// </summary>
		/// <param name="path">The path to the project to create the host for.</param>
		public SubSonicHost(string path)
		{
			_subsonic = CreateSubSonicProxy(path);
		}

		#endregion Constructors


		#region Properties

		/// <summary>
		/// Gets the hosted SubSonic proxy.
		/// </summary>
		public ISubSonic SubSonic
		{
			get { return _subsonic; }
		}

		#endregion Properties


		#region Methods

		/// <summary>
		/// Creates SubSonic in its own domain.
		/// </summary>
		/// <remarks>
		/// Assigns the configuration file with the highest priority to the newly created AppDomain.
		/// </remarks>
		/// <param name="path">The path to the project to create the generator for.</param>
		/// <returns>The SubSOnic interface.</returns>
		[MethodImpl(MethodImplOptions.NoInlining)] // Make sure the method is not inlined by the compiler.
		protected ISubSonic CreateSubSonicProxy(string path)
		{
			AppDomain domain = null;
			AppDomainSetup setup = new AppDomainSetup();
			string[] configFiles = /*ConfigurationProvider.*/FindProjectConfigFiles(path);

			AppDomain.CurrentDomain.AssemblyResolve += new ResolveEventHandler(CurrentDomain_AssemblyResolve);
			setup.ApplicationBase = Path.GetDirectoryName(Assembly.GetExecutingAssembly().Location);
			setup.ConfigurationFile = configFiles[0]; // Make System.Configuration.ConfigurationManager use this file in the new AppDomain.
			domain = AppDomain.CreateDomain(Guid.NewGuid().ToString(), null, setup);

			return (ISubSonic)domain.CreateInstanceAndUnwrap(_generatorType.Assembly.FullName, _generatorType.FullName, false, BindingFlags.CreateInstance | BindingFlags.Instance | BindingFlags.Public, null, new object[] { configFiles }, null, null, null);
		}

		//}
		/// <summary>
		/// Finds a project configuration files from a specified path.
		/// </summary>
		/// <remarks>
		/// The method starts by searching the given path, then goes up the directory hierarchy until 
		/// it finds the configuration files. It stops looking if a project, solution or the root
		/// directory is reached.
		/// </remarks>
		/// <param name="path">
		/// A file path or a directory path. 
		/// A directory path must end with Path.DirectorySeparatorChar.
		/// </param>
		/// <returns>The configuration files found.</returns>
		public string[] FindProjectConfigFiles(string path)
		{
			List<string> configFiles = new List<string>();

			string targetFile = null;
			string[] filesFound = null;
			string currentDirectory = Path.GetDirectoryName(path);

			// Format directory correctly.
			if (currentDirectory != string.Empty && !currentDirectory.EndsWith(Path.DirectorySeparatorChar.ToString()))
				currentDirectory = string.Format("{0}{1}", currentDirectory, Path.DirectorySeparatorChar);

			do
			{
				// Then look for Web.config
				targetFile = string.Format("{0}{1}", currentDirectory, "Web.config");
				if (File.Exists(targetFile))
					configFiles.Add(targetFile);

				// Then look for App.config
				targetFile = string.Format("{0}{1}", currentDirectory, "App.config");
				if (File.Exists(targetFile))
					configFiles.Add(targetFile);

				// If at least one config file has been found, stop.
				if (configFiles.Count != 0)
					break;

				// If we find a project file, stop.
				filesFound = Directory.GetFiles(currentDirectory, "*.*proj", SearchOption.TopDirectoryOnly);
				if (filesFound.Length != 0)
					break;

				// If we find a solution file, stop.
				filesFound = Directory.GetFiles(currentDirectory, "*.sln", SearchOption.TopDirectoryOnly);
				if (filesFound.Length != 0)
					break;

				// If we're at the root, stop.
				if (currentDirectory == string.Empty || currentDirectory == Path.GetPathRoot(path))
					break;

				// Else we move one directory up.
				currentDirectory = currentDirectory.Remove(currentDirectory.Length - 1); // Remove last Path.DirectorySeparatorChar
				currentDirectory = currentDirectory.Substring(0, currentDirectory.LastIndexOf(Path.DirectorySeparatorChar) + 1);
			}
			while (true);

			if (configFiles.Count == 0)
				throw new Exception(string.Format("No project configuration found from '{0}'.", path));

			return configFiles.ToArray();
		}

		#endregion Methods


		#region Event Handlers

		/// <summary>
		/// Occurs when the resolution of an assembly fails.
		/// </summary>
		/// <remarks>
		/// http://www.codeguru.com/forum/showthread.php?t=398030
		/// </remarks>
		/// <param name="sender">Source of the event.</param>
		/// <param name="args">Event parameters.</param>
		/// <returns>The System.Reflection.Assembly that resolves the type, assembly, or resource.</returns>
		[MethodImpl(MethodImplOptions.NoInlining)] // Make sure the method is not inlined by the compiler.
		Assembly CurrentDomain_AssemblyResolve(object sender, ResolveEventArgs args)
		{
			return Assembly.GetExecutingAssembly().FullName == args.Name ? Assembly.GetExecutingAssembly() : null;
		}

		#endregion Event Handlers
	}
}
