///////////////////////////////////////////////////////
// Code author: Martin Lapierre, http://devinstinct.com
///////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.IO;
using System.CodeDom;

namespace SubSonic.Tools
{
	/// <summary>
	/// Provides a single entry-point to SubSonic.
	/// </summary>
	/// <remarks>
	/// Used by SubSonicHost, who needs a MarshalByRefObject.
	/// An instance of this class is not thread safe.
	/// </remarks>
	public class SubSonicProxy : MarshalByRefObject, ISubSonic
	{
		#region Constructors

		/// <summary>
		/// Constructor.
		/// </summary>
		public SubSonicProxy()
		{
			DataService.LoadProviders(); // This is where SubSonic init its configuration... load it now.
		}

		/// <summary>
		/// Constructor.
		/// </summary>
		/// <param name="configFiles">The configuration files to use as override.</param>
		public SubSonicProxy(string[] configFiles)
		{
			// Hack to support relative path for code templates.
			// Only support current directory relativity (starting with "." or ".\").
			DataService.LoadProviders(); // This is where SubSonic init its configuration... load it now.
			if (!string.IsNullOrEmpty(CodeService.TemplateDirectory) && (CodeService.TemplateDirectory == "." || CodeService.TemplateDirectory.StartsWith(@".\")))
			{
				string directory = Path.GetDirectoryName(configFiles[0]);
				if (!directory.EndsWith(Path.DirectorySeparatorChar.ToString()))
					directory += Path.DirectorySeparatorChar;
				if (CodeService.TemplateDirectory == ".")
					CodeService.TemplateDirectory = directory;
				else
					CodeService.TemplateDirectory = directory + CodeService.TemplateDirectory.Substring(2);
			}
		}

		#endregion Constructors


		#region Methods

        private ICodeLanguage GetCodeLanguage(LanguageType languageType)
        {
            switch (languageType)
            {
                case LanguageType.VB:
                    return CodeLanguageFactory.GetByShortName(LanguageAbbreviation.VB);
                default:
                    return CodeLanguageFactory.GetByShortName(LanguageAbbreviation.C_SHARP);
            }
        }

		/// <summary>
		/// Generates the DAL in a single CodeCompileUnit from an input file.
		/// </summary>
		/// <param name="languageType">The language to generate the DAL for.</param>
		/// <returns>A single CodeCompileUnit.</returns>
		public CodeCompileUnit GenerateSingleUnit(LanguageType languageType)
		{
			return BuildService.GenerateSingleUnit(GetCodeLanguage(languageType));
		}

		/// <summary>
		/// Generates the DAL in multiple CodeCompileUnits for specific tables, views
		/// and, optionnaly, all stored procedures.
		/// </summary>
		/// <param name="languageType">The language to generate the DAL for.</param>
		/// <param name="tables">
		/// The list of tables to generate the DAL for; first element as "*" for all tables.
		/// Can be null or empty. A null value generates no table and no table struct. 
		/// <param name="views">
		/// The list of views to generate the DAL for; first element as "*" for all views.
		/// Can be null or empty. A null value generates no view and no view struct. 
		/// </param>
		/// <param name="sps">
		/// The list of stored procedures to generate the DAL for; first element as "*" for all stored procedures.
		/// Can be null or empty. A null value generates no stored procedures. 
		/// </param>
		/// <param name="providerName">
		/// The name of the provider to generate the DAL for.
		/// If null, generates code for all the providers.
		/// </param>
		/// <param name="addDependencies">Indicates whether to add the dependent tables or not.</param>
		/// <returns>A dictionary of suggested filenames/CodeCompileUnit pairs.</returns>
		public IDictionary<string, CodeCompileUnit> GenerateMultipleUnits(LanguageType languageType, string[] tables, string[] views, string[] sps, bool addDependencies, string providerName)
		{
            return BuildService.GenerateMultipleUnits(GetCodeLanguage(languageType), tables, views, sps, addDependencies, providerName);
		}

		/// <summary>
		/// Generates the DAL in multiple CodeCompileUnits from an input file.
		/// </summary>
		/// <param name="languageType">The language to generate the DAL for.</param>
		/// <param name="inputFile">The path to the input file.</param>
		/// <returns>A dictionary of suggested filenames/CodeCompileUnit pairs.</returns>
		public IDictionary<string, CodeCompileUnit> GenerateMultipleUnits(LanguageType languageType)
		{
			string[] tables = new string[] { "*" };
			string[] views = new string[] { "*" };
			string[] sps = new string[] { "*" };

			// Always generate tables & views; generate SP based on config file.
			return GenerateMultipleUnits(languageType, tables, views, sps, true, null);
		}

		/// <summary>
		/// Gets the list of tables.
		/// </summary>
		/// <param name="providerName">
		/// The name of the provider to get the tables for.
		/// If null, uses the default provider.
		/// </param>
		/// <returns>A list of table names.</returns>
		public string[] GetTables(string providerName)
		{
			if (providerName == null)
				return DataService.GetTableNames(DataService.Provider.Name);
			else
				return DataService.GetTableNames(providerName);
		}

		/// <summary>
		/// Gets the list of views.
		/// </summary>
		/// <param name="providerName">
		/// The name of the provider to get the views for.
		/// If null, uses the default provider.
		/// </param>
		/// <returns>A list of view names.</returns>
		public string[] GetViews(string providerName)
		{
			if (providerName == null)
				return DataService.GetViewNames(DataService.Provider.Name);
			else
				return DataService.GetViewNames(providerName);
		}

		/// <summary>
		/// Gets the list of stored procedures.
		/// </summary>
		/// <param name="providerName">
		/// The name of the provider to get the stored procedures for.
		/// If null, uses the default provider.
		/// </param>
		/// <returns>A list of stored procedure names.</returns>
		public string[] GetSPs(string providerName)
		{
			if (providerName == null)
				return DataService.GetSPList(DataService.Provider.Name);
			else
				return DataService.GetSPList(providerName);
		}

		/// <summary>
		/// Gets the list of provider names for the current application context.
		/// </summary>
		/// <returns>A list of provider names.</returns>
		public string[] GetProviderNames()
		{
			List<string> providerNames = new List<string>();
			foreach (DataProvider provider in DataService.Providers)
				providerNames.Add(provider.Name);
			return providerNames.ToArray();
		}

		#endregion Methods
	}
}
